--[[
	animation (анимация) - данные по смене изображения со временем, общие для объектов одного типа
	clip (клип) - инф-я о состоянии одной анимации для одного объекта
]]

FROM_START, BACKWARD = 0, 1

animation = {
	errors = {
		[0]="no error",
		[1]="name (parameter 1) is nil, empty or non-string value", [2]="image (parameter 2) is nil or non-userdata value", [3]="list (parameter 3) is nil, empty or non-table value", [4]="delay (parameter 6) is nil, non-positive number or non-table value",
		[5]="animation with this name already exist"
	},
}

-- список всех анимаций в игре (поименно)
-- [имя анимации] = { список клипов }
animation.list = { }

--[[ Создание новой анимации:
* animation.create(name, image, fx, fy, {list}, delay)
+	name - имя анимации
+	image - картинка, с кадрами анимации
+	fx,fy - размер одного кадра в пикселях
+	{list} - массив с координатами кадров в картинке через запятую (кол-во кадров = размер массива / 2)
+	delay - задержка между кадрами в мс (если массив, то в нем отдельная задержка после каж. кадра)
Возвращает код ошибки (и 0 если все получилось)
]]

--[[ ВНУТРЕННИЕ (внутренние, их не следует изменять другими функциями)
Поля анимаций (в скобках значение по умолчанию):
+	clips - список клипов, использующих эту анимацию
+	image - ссылка на изображение
+	fx,fy - размер одного кадра в пикселях
+	delays - массив с задержками между кадрами
+	frames - массив с координатами кадров анимации
+	num - кол-во кадров анимации
]]

--[[ Удаление анимации и всех ее кадров:
* animation.destroy(name)
	name - имя анимации (если анимация воспроизводится, то она останавливается)
]]

function animation.create(file, name, image, fx, fy, list, delay, events)
	-- проверяем типы осн. параметров
	if name == nil or type(name) ~= "string" then return 1 end
	if (animation.list[file] == nil) then animation.list[file] = { } end
	if animation.list[file][name] ~= nil then return 5 end
	if image == nil or type(image) ~= "userdata" then return 2 end
	if list == nil or type(list) ~= "table" or #list < 2 then return 3 end
	if delay == nil or (type(delay) ~= "table" and type(delay) ~= "number") or (type(delay) == "number" and delay <= 0) then return 4 end
	-- создаем пустую анимацию
	local anim,i = { },0
	anim.image = image
	anim.fx, anim.fy = fx or SPRITE_SIZE, fy or SPRITE_SIZE
	-- проверяем задержку
	anim.num = math.floor(#list/2)
	if type(delay) == "number" then
		local val = delay
		delay = { }
		for i = 1,anim.num,1 do
			delay[i] = val
		end
	end
	-- и список событий
	if (events ~= nil and type(delay) == "number") then
		local val = events
		events = { }
		for i = 1,anim.num,1 do
			events[i] = val
		end
	end
	-- записываем кадры и задержки между ними
	anim.frames, anim.delays = { }, { }
	for i = 1,anim.num,1 do
		anim.frames[i] = lg.newQuad(fx*list[2*i-1], fy*list[2*i], fx, fy, image:getWidth(), image:getHeight())
		anim.delays[i] = delay[i] or 1.0 --0.25
	end
	anim.events = events
	-- добавляем новую анимацию в таблицу в таблицу
	anim.clips = { }
	animation.list[file][name] = anim
	return 0
end

function animation.clone(new_file, old_file, image)
	if (animation.list[new_file] == nil) then animation.list[new_file] = { } end
	for k,v in pairs(animation.list[old_file]) do
		animation.list[new_file][k] = v
		animation.list[new_file][k].image = image
	end
end

local actor_list

function animation.update(dt)
	for id,obj in pairs(object.list) do
		if (obj.clip ~= nil) then
			clip.update(obj.clip, dt)
		end
	end
end

function animation.destroy(name)
end

--[[ Создание нового клипа:
* clip.create(name, obj, speed, repeating)
+	name - имя анимации, воспроизводимой клипом
+	obj - к какому объекту привязан клип (нужно для рисовании анимации при воспроизведении)
+	speed - скорость воспроизведения
+	repeating - повторение анимации
Возвращает ссылку на клип анимации
]]

--[[ Анимации внутри объекта:
	clips - массив всех анимаций
	clip - текущая воспр-ся анимация
	clip_index - индекс текущей анимации (nil, если анимация не воспроизводится)
]]

--[[ ВНЕШНИЕ (для управления клипами из других функций используется только имя анимации)
Функции:
+	Play() - продолжить анимацию (если onRepeat == false, то при завершении анимации она остановится на первом кадре)
+	Rewind(N) - перемотать N-кадр или на начало
+	Stop() - остановить анимацию
+	Repeat(true|false) - делает анимацию цикличной
	Speed - установить скорость анимации
Проверка состояния (что возвращает):
	isPlaying - воспроизводится?
	isRepeat - зациклена?
	getSpeed - возвращает скорость
	curFrame - номер текущего кадра
	curDelay - сколько мс осталось до след. кадра
]]

--[[ ВНУТРЕННИЕ
+	frame - номер текущего кадра анимации
+	delay - текущая оставшаяся задержка до след. кадра
	playing (false) - воспроизводится?, если false, то анимация остановлена
	speed (1) - скорость (<0 - для обратного воспроизведения, 0 означает невозможность воспр-я)
+	onRepeat (false) - воспроизводить анимацию заново при достижении посл. кадра?
]]

clip = { }

local function add_clip(self, obj, index)
	if obj.clips == nil then obj.clips = { } end
	--if (obj.clips[index] == nil) then	-- не добавляем анимацию, если она уже есть
		obj.clips[index] = self
		return true
	--end
	--return false
end

function clip.create(file, name, obj, speed, repeating, index, playback)
	if animation.list[file] == nil or animation.list[file][name] == nil then return nil; end
	local self = { }
	self.anim = animation.list[file][name]
	self.frame = 1
	self.delay = animation.list[file][name].delays[1]
	self.speed = speed or 1.0
	self.playing = false
	self.onRepeat = repeating or false
	self.playback = playback or 0
	self.obj = obj
	--
	if add_clip(self, obj, index or 0) then return self end
	return nil
end

local quad, viewport

function clip.show_frame(self, index)	-- клип, номер кадра (если номер не указан, то показывается текущий)
	if (index == nil) then index = self.frame end
	self.delay = self.anim.delays[index]
	self.obj.sprite = { self.anim.image, self.anim.frames[index] }
	if (self.anim.events ~= nil and self.anim.events[index] ~= nil) then	
		if (self.obj.clip_event ~= nil) then self.obj.clip_event(self.obj, index, self.anim.events[index]) end
	end
end

local function stopClip(self)
	if self.obj.clip ~= nil then
		self.obj.clip = nil
		self.obj.clip_index = nil
	end
end

local function rewindClip(self)
	self.frame = 1
end

function clip.update(self, dt) -- исп-ся локальное поле anim
	-- уменьшаем таймер до след кадра
	--self.delay = self.delay - math.abs(self.speed)*dt
	self.delay = self.delay - math.abs(self.speed * self.obj.speed)*dt
	-- переходим к след. кадру
	if (self.delay < 0) then
		-- воспроизведение вперед
		if (self.speed > 0) then
			if self.frame == self.anim.num then -- достигли конца анимации
				if self.onRepeat then
					-- повторяем с начала или в обр. сторону
					if (self.playback == FROM_START) then self.frame = 1 end
					if (self.playback == BACKWARD) then self.frame = #self.anim.frames; self.speed = -self.speed end
				else
					stopClip(self); rewindClip(self); return	-- останавливаем клип
				end
			else
				self.frame = self.frame + 1
			end
			if (self.obj.playback_event ~= nil) then self.obj.playback_event(self.obj, self.obj.clip_index) end
		end
		-- воспроизведение назад
		if (self.speed < 0) then
			if self.frame == 1 then -- достигли начала анимации
				if self.onRepeat then
					if (self.playback == FROM_START) then self.frame = #self.anim.frames end
					if (self.playback == BACKWARD) then self.frame = 1; self.speed = -self.speed end
				else
					stopClip(self); rewindClip(self); return	-- останавливаем клип
				end
			else
				self.frame = self.frame - 1
			end		
			if (self.obj.playback_event ~= nil) then self.obj.playback_event(self.obj, self.obj.clip_index) end
		end
		clip.show_frame(self)
	end
end

local function play(self, index, frame) -- self - это клип
	self.obj.clip = self
	self.obj.clip_index = index
	self.frame = frame or 1
	clip.show_frame(self, frame or 1)
end

local played_clip

function clip.Play(obj, index, frame)
	if (obj.clips == nil) then return -1 end
	played_clip = obj.clips[index] 
	if (played_clip == nil) then return -1 end
	if (played_clip.speed > 0) then
		play(played_clip, index, (frame or 1));
	else
		play(played_clip, index, (frame or #played_clip.anim.frames));
	end
end

function clip.Stop(obj) -- останавливаем указанный текущий клип указанного объекта
	if obj ~= nil and obj.clip ~= nil then
		obj.clip = nil
		obj.clip_index = nil
	end	
end

function clip.Rewind(obj, index, frame)
	if obj == nil then return end
	if frame == nil then frame = 1 end
	if (index == nil) then
		obj.clip.frame = frame
	else
		obj.clips[index].frame = frame
	end
end

function clip:Repeat(r)
	self.onRepeat = r
end

function clip.SetRepeat(obj, index, r)
	obj.clips[index].onRepeat = r
end

function clip.SetSpeed(obj, index, value)
	obj.clips[index].speed = value
end

function clip.getSpeed(obj, index)
	return obj.clips[index].speed
end

function clip:destroy(self)
	--[[
	local name = self.name
	local list = animation.list[name].clips
	for i = 1,list,1 do
		if list[i] == self then table.remove(animation.list[name].clips, i); break; end
	end
	--]]
end

function clip:get_num_frames()
	return #self.anim.frames
end


-- ДОПОЛНИТЕЛЬНЫЕ ФУНКЦИИ

function clip.IsPlayingAnim(obj)
	return obj.clip ~= nil
end

-- проверка - воспроизводится ли сейчас у obj анимация с индексом index?
-- если индекс не указан то проверяется сам факт воспр-я любой анимации
function clip.IsPlaying(obj, index)
	if (obj == nil) then return end
	if (index == nil) then
		return obj.clip ~= nil
	else
		return obj.clip == obj.clips[index]
	end
end

function clip.GetPlayingIndex(obj)
	return obj.clip_index
end

-- сменить спрайт на кадр анимации, не запуская саму анимацию
function clip.ShowFrame(obj, index, frame)
	if (obj.clips == nil) then return -1 end
	played_clip = obj.clips[index]
	if (played_clip == nil) then return -1 end
	clip.show_frame(played_clip, frame or 1)
end

-- возвращает номер текущего кадра в анимации index
function clip.get_frame(obj, index)
	if obj == nil then return 0 end
	if (index == nil) then
		return obj.clip.frame
	else
		return obj.clips[index].frame
	end
end

-- сообщает кол-во кадров в анимации
function clip.GetNumFrames(obj, index)
	if (obj.clips == nil) then return -1 end
	played_clip = obj.clips[index]
	if (played_clip == nil) then return -1 end
	return clip.get_num_frames(played_clip)
end
