--FILE, BACK, MIDDLE, FRONT = 0, 1, 2, 3		-- layers

object = {
	list = { }, root = { },
	layer = { [0]={}, [1]={}, [2]={}, [3]={} },
}
object.__index = object

local last_id = 0

local function next_id()
	last_id = last_id + 1
	return last_id
end

function object.create(x,y,layer)
	-- создаем новый объект
	local self = setmetatable({}, object)
	self.x, self.y = x, y
	self.name = ""
	self.v = { x = 0, y = 0 }; -- скорость движения (пикселей в секунду)
	self.speed = 0
	self.update = nil
	self.layer = layer or 2 -- MIDDLE
	-- родительские связи
	self.parent = nil
	-- записываем в список
	self.id = next_id()
	object.list[self.id] = self		-- все объекты
	object.root[self.id] = self		-- иерархия (родитель <= ребенки)
	if (use_queue) then camera.add_queue(self) else object.layer[self.layer][self.id] = self	end -- порядок рисования
	return self
end

function object.attach(self,parent)
	if self == parent or self.parent == parent then return nil end
	if self.parent ~= nil then object.detach(self) end
	self.parent = parent
	if parent.childs == nil then parent.childs = { } end
	parent.childs[self.id] = self
	object.root[self.id] = nil
end

function object.detach(self)
	if self.parent ~= nil then
		self.parent.childs[self.id] = nil
		if #self.parent.childs == 0 then self.parent.childs = nil end
		self.parent = nil
		object.root[self.id] = self
	end
end

-- удаляем одинокий объект
function destroy_components(self)
	-- своя функция уничтожения
	if (self.destroy ~= nil) then self.destroy(self) end
	-- удаляем все компоненты
	if (self.typ ~= nil and self.typ == "actor") then actor.destroy(self) end
	if (self.is_trigger) then trigger.destroy(self) end
	if (self.phys ~= nil) then collision.destroy(self); end
	if (self.shape ~= nil) then
		self.shape.obj = nil
		fizz.removeShape(self.shape)
	end
	-- удаляем сам объект
	if (use_queue) then camera.remove_queue(self) else object.layer[self.layer][self.id] = nil end
	object.list[self.id] = nil
	object.root[self.id] = nil
	self = nil
end

function object:destroy()
	self.on_destroy = true
end

function object:execute_all_destroy()
	for id,obj in pairs(object.list) do
		if obj.on_destroy then object.execute_destroy(obj) end
	end
end

-- удаление объекта вместе с дочерними
function object:execute_destroy()
	-- сначала удаляем детей
	if self.childs ~= nil then
		for index,obj in pairs(self.childs) do
			self.childs[obj.id] = nil
			obj.parent = nil
			object.execute_destroy(obj)
		end
	end
	-- потом себя
	destroy_components(self)
end

function object.destroy_all()
	for id,obj in pairs(object.list) do
		if not obj.dont_destroy then object.execute_destroy(obj) end
	end
end

function object:change_layer(val)
	if (use_queue) then
		camera.remove_queue(self)
		self.layer = val
		camera.add_queue(self)
	else
		object.layer[val][self.id] = self
		object.layer[self.layer][self.id] = nil
		self.layer = val
	end
end

function object.find_first(name)
	for i,cur_obj in pairs(object.list) do
		if cur_obj.name == name then return cur_obj; end
	end
	return nil
end


-- MOVING and UPDATE

function object:teleport(new_x, new_y)
	self.x, self.y = new_x, new_y
	fizz.setPosition(self.shape, new_x, new_y)
end

function object:move(vect, val)
	vector.norm(vect)
	self.v.x, self.v.y = vect.x, vect.y
	self.speed = val or 1
end

function object:move_to(x, y, val)
	local vect = { x = x-self.x, y = y-self.y }
	vector.norm(vect)
	self.v.x, self.v.y = vect.x, vect.y
	self.speed = val or 1
end

function object:stop()
	self.v.x, self.v.y = 0, 0
end

function object:update_pos(dt)
	self.x = self.x + self.v.x * (self.speed or 1) * dt
	self.y = self.y + self.v.y * (self.speed or 1) * dt
end

prev = { x = 0, y = 0 }

function object.update_all(dt)
	for i,obj in pairs(object.root) do
		prev.x, prev.y = obj.x, obj.y
		-- потом обновляем себя
		if (obj.update ~= nil) then obj.update(obj, dt) end
		-- обновляем сперва дочерние объекты
		if obj.childs ~= nil then
			for i,child in pairs(obj.childs) do
				-- двигаем объект вместе с родителем
				child.x = child.x + (obj.x - prev.x)
				child.y = child.y + (obj.y - prev.y)
				-- обновляем
				if (child.update ~= nil) then child.update(child, dt) end
			end
		end
	end
end

local sprite

function object:normal_draw()
	lg.draw(self.sprite[1], self.sprite[2], self.x, self.y, 0, 1, 1, self.ox, self.oy)
	if (is_debug) then lg.points(self.x, self.y) end
end

function object:draw_image()
	if not self.hide then
		sprite = self.sprite
		lg.draw(sprite[1], sprite[2], self.x, self.y, 0, 1, 1, self.ox, self.oy)
	end
end

local blendMode, prev_color

function object.real_draw(obj)		-- мигание
	if (obj.is_blink) then
		prev_color = { lg.getColor() }
			lg.setColor(color.green2)
			obj:draw()
		lg.setColor(prev_color)
	else
		obj:draw()
	end
end
-- [[
function object.draw_layer(index)
	for i,obj in pairs(object.layer[index]) do
		object.real_draw(obj)
	end
end

function object.draw_all()
	for l = 0,#object.layer,1 do
		for i,obj in pairs(object.layer[l]) do
			object.real_draw(obj)
		end
	end
end

function object.post_draw_all()
	for l = 0,#object.layer,1 do
		for i,obj in pairs(object.layer[l]) do
			if (obj.post_draw ~= nil) then obj:post_draw() end
		end
	end
end

-- ОТЛАДКА

function object.print_list()
	debuginfo.text = ""
	for id,obj in pairs(object.list) do
		debuginfo.text = debuginfo.text..obj.name.."\n"
	end
end